<?php

namespace Drupal\charts_echarts\Plugin\chart\Library;

use Drupal\charts\Plugin\chart\Library\ChartBase;
use Drupal\Component\Utility\Color;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Url;

/**
 * Define a concrete class for a Chart.
 *
 * @Chart(
 *   id = "echarts",
 *   name = @Translation("eCharts.js")
 * )
 */
class Echarts extends ChartBase {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['placeholder'] = [
      '#title' => $this->t('Placeholder'),
      '#type' => 'fieldset',
      '#description' => $this->t(
        'This is a placeholder for eChart.js-specific library options. If you would like to help build this out, please work from <a href="@issue_link">this issue</a>.', [
          '@issue_link' => Url::fromUri('https://www.drupal.org/project/charts/issues/3046984')
            ->toString(),
        ]),
    ];
    $xaxis_configuration = $this->configuration['xaxis'] ?? [];
    $yaxis_configuration = $this->configuration['yaxis'] ?? [];
    $form['xaxis'] = [
      '#title' => $this->t('X-Axis Settings'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
    ];
    $form['xaxis']['autoskip'] = [
      '#title' => $this->t('Enable autoskip'),
      '#type' => 'checkbox',
      '#default_value' => $xaxis_configuration['autoskip'] ?? 1,
    ];
    $form['xaxis']['horizontal_axis_title_align'] = [
      '#title' => $this->t('Align horizontal axis title'),
      '#type' => 'select',
      '#options' => [
        'start' => $this->t('Start'),
        'center' => $this->t('Center'),
        'end' => $this->t('End'),
      ],
      '#default_value' => $xaxis_configuration['horizontal_axis_title_align'] ?? '',
    ];
    $form['yaxis'] = [
      '#title' => $this->t('Y-Axis Settings'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
    ];
    $form['yaxis']['vertical_axis_title_align'] = [
      '#title' => $this->t('Align vertical axis title'),
      '#type' => 'select',
      '#options' => [
        'start' => $this->t('Start'),
        'center' => $this->t('Center'),
        'end' => $this->t('End'),
      ],
      '#default_value' => $yaxis_configuration['vertical_axis_title_align'] ?? '',
    ];

    return $form;
  }

  /**
   * Build configurations.
   *
   * @param array $form
   *   The form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['xaxis'] = $values['xaxis'];
      $this->configuration['yaxis'] = $values['yaxis'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function preRender(array $element) {
    $chart_definition = [];

    if (!isset($element['#id'])) {
      $element['#id'] = Html::getUniqueId('echarts-render');
    }

    $chart_definition = $this->populateDatasets($element, $chart_definition);
    $chart_definition = $this->populateOptions($element, $chart_definition);
    //\Drupal::messenger()->addMessage(' $chart_definition >> '. json_encode($chart_definition));

    $element['#attached']['library'][] = 'charts_echarts/echarts';
    $element['#attributes']['class'][] = 'charts-echarts';
    $element['#chart_definition'] = $chart_definition;

    return $element;
  }

  /**
   * Populate options.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  private function populateOptions(array $element, array $chart_definition) {
    $chart_type = $this->populateChartType($element);
    $children = Element::children($element);

    /*
     * Setting defaults based on what Views uses. However, API users may
     * have different keys for their X and Y axes.
     */
    $x_axis_key = 'xaxis';
    $x_axis_key_cap = 'xAxis';
    $y_axis_key = 'yaxis';
    $y_axis_key_cap = 'yAxis';
    $categories = [];
    foreach ($children as $child) {
      $type = $element[$child]['#type'];
      if ($type === 'chart_xaxis') {
        $x_axis_key = $child;
        $x_axis_key_cap = ($child === 'xaxis') ? 'xAxis' : $child;
        $categories = array_map('strip_tags', $element[$child]['#labels']);
        // Merge in axis raw options.
        if (!empty($element[$child]['#raw_options'])) {
          $categories = NestedArray::mergeDeepArray([$element[$child]['#raw_options'], $categories,]);
        }
        $chart_definition['options'][$x_axis_key_cap]['data'] = $categories;
      }
      if ($type === 'chart_yaxis') {
        $y_axis_key = $child;
        $y_axis_key_cap = ($child === 'yaxis') ? 'yAxis' : $child;;
      }
    }

    $xaxis_configuration = $this->configuration['xaxis'] ?? [];
    $yaxis_configuration = $this->configuration['yaxis'] ?? [];

    if (!in_array($chart_type, ['pie', 'doughnut'])) {
      if (!empty($element['#stacking']) && $element['#stacking'] == 1) {
        $stacking = TRUE;
      }
      else {
        $stacking = FALSE;
      }
      if ($chart_type !== 'radar') {

        //@TODO
        //$chart_definition['options'][$x_axis_key] - rotation & stacking

        if (!empty($element[$y_axis_key]['#min'])) {
          $chart_definition['options'][$y_axis_key_cap]['min'] = $element[$y_axis_key]['#min'];
        }
        if (!empty($element[$y_axis_key]['#max'])) {
          $chart_definition['options'][$y_axis_key_cap]['max'] = $element[$y_axis_key]['#max'];
        }
        if (!empty($element[$y_axis_key]['#title'])) {
          $chart_definition['options'][$y_axis_key_cap]['name'] = $element[$y_axis_key]['#title'];
          $chart_definition['options'][$y_axis_key_cap]['nameLocation'] = 'middle';
          $chart_definition['options'][$y_axis_key_cap]['nameGap'] = '45';
        }
        if (!empty($element[$x_axis_key]['#title'])) {
          $chart_definition['options'][$x_axis_key_cap]['name'] = $element[$x_axis_key]['#title'];
          $chart_definition['options'][$x_axis_key_cap]['nameLocation'] = 'middle';
          $chart_definition['options'][$x_axis_key_cap]['nameGap'] = '45';
        }

      }
    }

    $chart_definition['options']['title'] = $this->buildTitle($element);
    $chart_definition['options']['tooltip'] = $this->buildTooltip($element);
    $chart_definition['options']['toolbox'] = $this->buildToolbox($element);
    $chart_definition['options']['legend'] = $this->buildLegend($element);



    // Merge in chart raw options.
    if (!empty($element['#raw_options'])) {
      $chart_definition = NestedArray::mergeDeepArray([
        $chart_definition,
        $element['#raw_options'],
      ]);
    }

    //@TODO
    //add Grid to configuration form
    /**
    $chart_definition['options']['grid']['top'] = 6;
    $chart_definition['options']['grid']['right'] = 0;
    $chart_definition['options']['grid']['bottom'] = 50;
    $chart_definition['options']['grid']['left'] = 45;
    $chart_definition['options']['grid']['width'] = 'auto';
    $chart_definition['options']['grid']['height'] = 'auto';
     */

    return $chart_definition;
  }

  /**
   * Populate categories.
   *
   * @param array $element
   *   The element.
   *
   * @return array
   *   Return the chart definition.
   */
  private function populateCategories(array $element) {
    $children = Element::children($element);
    $categories = [];
    foreach ($children as $child) {
      $type = $element[$child]['#type'];
      if ($type === 'chart_xaxis') {
        $categories = array_map('strip_tags', $element[$child]['#labels']);
        // Merge in axis raw options.
        if (!empty($element[$child]['#raw_options'])) {
          $categories = NestedArray::mergeDeepArray([$element[$child]['#raw_options'], $categories,]);
        }
      }
    }

    return $categories;
  }

  /**
   * Populate Dataset.
   *
   * @param array $element
   *   The element.
   * @param array $chart_definition
   *   The chart definition.
   *
   * @return array
   *   Return the chart definition.
   */
  private function populateDatasets(array $element, array $chart_definition) {
    $chart_type = $this->populateChartType($element);
    $datasets = [];
    foreach (Element::children($element) as $key) {
      if ($element[$key]['#type'] === 'chart_data') {
        $series_data = [];
        $dataset = new \stdClass();
        // Populate the data.
        foreach ($element[$key]['#data'] as $data_index => $data) {

          if (isset($series_data[$data_index])) {
            $series_data[$data_index][] = $data;
          }
          else {
            if ($chart_type === 'scatter') {
              $data = ['y' => $data[1], 'x' => $data[0]];
            }
            /*
             * This is here to account for differences between Views and
             * the API. Will change if someone can find a better way.
             */
            if(in_array($chart_type, ['pie','doughnut', 'gauge']) && !empty($data[1])) {
              $data = ['value' => $data[1], 'name' => $data[0]];
            }
            $series_data[$data_index] = $data;
          }
        }

        if (!empty($element['#stacking']) && $element['#stacking'] == 1) {
          $dataset->stack = 'total';
        }
        $dataset->name = $element[$key]['#title'];
        $dataset->data = $series_data;
        if (!in_array($chart_type, ['pie', 'doughnut'])) {
          //@TODO
          //$dataset->borderColor = $element[$key]['#color'];
        }

        $series_type = isset($element[$key]['#chart_type']) ? $this->populateChartType($element[$key]) : $chart_type;
        $dataset->type = $series_type;

        if (in_array($chart_type, ['pie', 'doughnut'])) {
          $dataset->type = 'pie';
          if ($chart_type === 'pie') {
            $dataset->radius = '60%';
          }
          if ($chart_type === 'doughnut') {
            $dataset->radius = ['40%', '70%'];
            //$dataset->avoidLabelOverlap = FALSE;
          }
        }

        //@TODO
        //add to configuration form
        if ($chart_type === 'gauge') {
          $dataset->axisLine->lineStyle->width = 30;
          $dataset->axisLine->lineStyle->color = [
            [0.3, '#67e0e3'],
            [0.7, '#37a2da'],
            [1, '#fd666d'],
          ];
          $dataset->axisTick->distance = -30;
          $dataset->axisTick->length = 8;
          $dataset->axisTick->lineStyle->color = '#fff';
          $dataset->axisTick->lineStyle->width = 2;

          $dataset->splitLine->distance = -30;
          $dataset->splitLine->length = 30;
          $dataset->splitLine->lineStyle->color = '#fff';
          $dataset->splitLine->lineStyle->width = 4;

          $dataset->axisLabel->color = 'auto';
          $dataset->axisLabel->distance = 40;
          $dataset->axisLabel->fontSize = 14;
        }

        $dataset->backgroundStyle->color = $element[$key]['#color'];

        //@TODO
        //Investigate different charts backgroundStyle & borderColor
        if (!empty($element[$key]['#chart_type']) && $element[$key]['#chart_type'] === 'area') {
          $dataset->backgroundStyle->color = $this->getTranslucentColor($element[$key]['#color']);
        }
        elseif ($element['#chart_type'] === 'area') {
          $dataset->backgroundStyle->color = $this->getTranslucentColor($element[$key]['#color']);
          $dataset->areaStyle = [''];
        }
        else {
          //$dataset->fill = FALSE;
        }

        $itemStyle['normal']['barBorderRadius'] = [2, 2, 0, 0];
        $itemStyle['normal']['barBorderWidth'] = [2];
        $dataset->itemStyle = $itemStyle;

        $datasets[] = $dataset;
      }

      // Merge in axis raw options.
      if (!empty($element[$key]['#raw_options'])) {
        $datasets = NestedArray::mergeDeepArray([
          $datasets,
          $element[$key]['#raw_options'],
        ]);
      }
    }

    $chart_definition['options']['series'] = $datasets;
    //\Drupal::messenger()->addMessage(' $chart_definition >> '. json_encode($chart_definition));
    return $chart_definition;
  }

  /**
   * Outputs a type that can be used by Chart.js.
   *
   * @param array $element
   *   The given element.
   *
   * @return string
   *   The generated type.
   */
  protected function populateChartType(array $element) {
    switch ($element['#chart_type']) {
      case 'bar':

      case 'column':
        $type = 'bar';
        break;

      case 'area':

      case 'spline':
        $type = 'line';
        break;

      case 'donut':
        $type = 'doughnut';
        break;

      case 'gauge':
        $type = 'gauge';
        break;

      default:
        $type = $element['#chart_type'];
        break;
    }
    if (isset($element['#polar']) && $element['#polar'] == 1) {
      $type = 'radar';
    }

    return $type;
  }

  /**
   * Builds legend based on element properties.
   *
   * @param array $element
   *   The element.
   *
   * @return array
   *   The legend array.
   */
  protected function buildLegend(array $element) {
    $chart_type = $this->populateChartType($element);
    $legend = [];
    // Configure the legend display.
    $legend['show'] = (bool) $element['#legend'];

    if (in_array($chart_type, ['pie', 'doughnut'])) {
      $legend['orient'] = 'vertical';
      $legend['left'] = $element['#legend_position'];
      $legend['top'] = 'bottom';
    }
    else {
      $legend['orient'] = 'horizontal';
      $legend[$element['#legend_position']] = 0;
    }
    $legend['type'] = 'scroll';

    // Configure legend position.
    if (!empty($element['#legend_position'])) {
      if (!empty($element['#legend_font_weight'])) {
        $legend['textStyle']['fontWeight'] = $element['#legend_font_weight'];
      }
      if (!empty($element['#legend_font_style'])) {
        $legend['textStyle']['fontStyle'] = $element['#legend_font_style'];
      }
      if (!empty($element['#legend_font_size'])) {
        $legend['textStyle']['fontSize'] = $element['#legend_font_size'];
      }
    }
    //\Drupal::messenger()->addMessage(' $legend >> '. json_encode($legend));
    return $legend;
  }

  /**
   * Builds tooltip based on element properties.
   *
   * @param array $element
   *   The element.
   *
   * @return array $tooltip
   *   The tooltip array.
   */
  protected function buildTooltip(array $element) {
    $chart_type = $this->populateChartType($element);
    // Configure the tooltip display.
    $tooltip = [];
    $tooltip['show'] = (bool)$element['#tooltips'];

    if (in_array($chart_type, ['pie', 'doughnut'])) {
      $tooltip['trigger'] = 'item';
    }
    else {
      $tooltip['trigger'] = 'axis';
      $tooltip['showContent'] = TRUE;
      $tooltip['alwaysShowContent'] = TRUE;
      $tooltip['axisPointer']['label']['show'] = FALSE;
    }

    $tooltip['triggerOn'] = 'mousemove';
    $tooltip['textStyle']['fontSize'] = '10';

    return $tooltip;
  }

  /**
   * Builds toolbox based on element properties.
   *
   * @param array $element
   *   The element.
   *
   * @return array $tooltip
   *   The toolbox array.
   */
  protected function buildToolbox(array $element) {
    $toolbox = [];
    // Configure the tooltip display.
    $toolbox['show'] = TRUE;
    $toolbox['feature']['mark']['show'] = TRUE;
    $toolbox['feature']['dataZoom']['yAxisIndex'] = "none";
    $toolbox['feature']['magicType']['show'] = TRUE;
    $toolbox['feature']['magicType']['type'] = ["line", "bar"];
    $toolbox['feature']['magicType']['title']['line'] = "Line";
    $toolbox['feature']['magicType']['title']['bar'] = "Bar";
    $toolbox['feature']['restore']['show'] = TRUE;
    $toolbox['feature']['restore']['title'] = "Reset";
    $toolbox['feature']['saveAsImage']['show'] = TRUE;
    $toolbox['feature']['saveAsImage']['title'] = "Save";


    return $toolbox;
  }

  /**
   * Builds title based on element properties.
   *
   * @param array $element
   *   The element.
   *
   * @return array $title
   *   The title array.
   */
  protected function buildTitle(array $element) {
    $title = [];
    if (!empty($element['#title'])) {
      $title = [
        'show' => TRUE,
        'text' => $element['#title'],
      ];
      if (!empty($element['#title_position'])) {
        if (in_array($element['#title_position'], ['in', 'out'])) {
          $title['top'] = 'auto';
        }
        else {
          $title[$element['#title_position']] = 'auto';
        }
      }
      if (!empty($element['#title_color'])) {
        $title['textStyle']['color'] = $element['#title_color'];
      }
      if (!empty($element['#title_font_weight'])) {
        $title['textStyle']['fontWeight'] = $element['#title_font_weight'];
      }
      if (!empty($element['#title_font_style'])) {
        $title['textStyle']['fontStyle'] = $element['#title_font_style'];
      }
      if (!empty($element['#title_font_size'])) {
        $title['textStyle']['fontSize'] = $element['#title_font_size'];
      }
    }

    return $title;
  }

  /**
   * Get translucent color.
   *
   * @param string $color
   *   The color.
   *
   * @return string
   *   The color.
   */
  protected function getTranslucentColor($color) {
    if (!$color) {
      return '';
    }
    $rgb = Color::hexToRgb($color);

    return 'rgba(' . implode(",", $rgb) . ',' . 0.5 . ')';
  }

}
